import shutil
import pytest

from pyreindexer import RxConnector
from tests.helpers.index import *
from tests.helpers.items import *
from tests.helpers.log_helper import log_fixture
from tests.helpers.metadata import put_metadata
from tests.helpers.namespace import *
from tests.test_data.constants import index_definition, item_definition


def pytest_addoption(parser):
    parser.addoption("--mode", action="store", default='builtin', help='builtin or cproto')


@pytest.fixture(scope="session", autouse=True)
def log_setup(request):
    """ Execute once before test run
    """
    log_fixture.info("Work with pyreindexer connector using {} mode".format(request.config.getoption("--mode")))


@pytest.fixture(scope="session")
def database(request):
    """
    Create a database
    """
    mode = request.config.getoption('--mode')
    db_name = 'test_db'
    if mode == 'builtin':
        prefix = 'builtin://tmp/'
    elif mode == 'cproto':
        prefix = 'cproto://127.0.0.1:6534/'
    else:
        raise ConnectionError
    db = RxConnector(prefix + db_name)
    yield db, db_name
    db.close()
    shutil.rmtree('tmp/', ignore_errors=True)


@pytest.fixture(scope="function")
def namespace(database):
    """
    Create a namespace
    """
    db, db_name = database
    ns_name = 'new_ns'
    create_namespace(database, ns_name)
    yield db, ns_name
    drop_namespace(database, ns_name)


@pytest.fixture(scope="function")
def index(namespace):
    """
    Create an index to namespace
    """
    create_index(namespace, index_definition)
    yield
    drop_index(namespace, 'id')


@pytest.fixture(scope="function")
def item(namespace):
    """
    Create an item to namespace
    """
    insert_item(namespace, item_definition)
    yield item_definition
    delete_item(namespace, item_definition)


@pytest.fixture(scope="function")
def items(namespace):
    """
    Create items to namespace
    """
    for i in range(10):
        insert_item(namespace, {"id": i+1, "val": "testval" + str(i+1)})
    yield
    for i in range(10):
        delete_item(namespace, {"id": i+1, "val": "testval" + str(i+1)})


@pytest.fixture(scope="function")
def metadata(namespace):
    """
    Put metadata  to namespace
    """
    key, value = 'key', 'value'
    put_metadata(namespace, key, value)
    yield key, value


@pytest.fixture(scope="function")
def second_namespace_for_join(database):
    db, db_name = database
    second_namespace_name = 'test_ns_for_join'
    db.namespace_open(second_namespace_name)
    db.index_add(second_namespace_name, index_definition)
    second_ns_item_definition = {"id": 100, "second_ns_val": "second_ns_testval"}
    second_ns_item_definition_join = {"id": 1, "second_ns_val": "second_ns_testval_1"}
    db.item_insert(second_namespace_name, second_ns_item_definition)
    db.item_insert(second_namespace_name, second_ns_item_definition_join)
    yield second_namespace_name, second_ns_item_definition_join



from tests.helpers.log_helper import log_operation


def create_index(namespace, index_def):
    """
    Create an index
    """
    db, namespace_name = namespace
    log_operation.info(f"Create an index to namespace '{namespace_name}', index={index_def}")
    db.index_add(namespace_name, index_def)


def update_index(namespace, index_def):
    """
    Update an index
    """
    db, namespace_name = namespace
    log_operation.info(f"Update an index to namespace '{namespace_name}', new index={index_def}")
    db.index_update(namespace_name, index_def)


def drop_index(namespace, index_name):
    """
    Drop index from namespace
    """
    db, namespace_name = namespace
    log_operation.info(f"Drop index from namespace '{namespace_name}', index name = '{index_name}'")
    db.index_drop(namespace_name, index_name)



from tests.helpers.log_helper import log_operation


def put_metadata(namespace, key, value):
    db, namespace_name = namespace
    log_operation.info(f"Put metadata '{key}: {value}' to namespace '{namespace_name}'")
    db.meta_put(namespace_name, key, value)
    return key, value


def get_metadata_keys(namespace):
    db, namespace_name = namespace
    log_operation.info("Get list of metadata keys")
    meta_list = db.meta_enum(namespace_name)
    return meta_list


def get_metadata_by_key(namespace, key):
    db, namespace_name = namespace
    value = db.meta_get(namespace_name, key)
    log_operation.info(f"Get metadata value by key: '{key}: {value}'")
    return value



index_definition = {
   "name": "id",
   "field_type": "int",
   "index_type": "hash",
   "is_pk": True,
   "is_array": False,
   "is_dense": False,
   "is_sparse": False,
   "collate_mode": "none",
   "sort_order_letters": "",
   "expire_after": 0,
   "config": {

   },
   "json_paths": [
      "id"
   ]
}

updated_index_definition = {
   "name": "id",
   "field_type": "int64",
   "index_type": "hash",
   "is_pk": True,
   "is_array": False,
   "is_dense": False,
   "is_sparse": False,
   "collate_mode": "none",
   "sort_order_letters": "",
   "expire_after": 0,
   "config": {

   },
   "json_paths": [
      "id_new"
   ]
}

special_namespaces = [{"name": "#namespaces"},
        {"name": "#memstats"},
        {"name": "#perfstats"},
        {"name": "#config"},
        {"name": "#queriesperfstats"},
        {"name": "#activitystats"},
        {"name": "#clientsstats"}]

special_namespaces_cluster = [{"name": "#namespaces"},
        {"name": "#memstats"},
        {"name": "#perfstats"},
        {"name": "#config"},
        {"name": "#queriesperfstats"},
        {"name": "#activitystats"},
        {"name": "#clientsstats"},
        {"name": "#replicationstats"}]

item_definition = {'id': 100, 'val': "testval"}




from hamcrest import *

from tests.helpers.metadata import *


class TestCrudMetadata:
    def test_initial_namespace_has_no_metadata(self, namespace):
        # Given("Create namespace")
        # When ("Get list of metadata keys")
        meta_list = get_metadata_keys(namespace)
        # Then ("Check that list of metadata is empty")
        assert_that(meta_list, has_length(0), "Metadata is not empty")
        assert_that(meta_list, equal_to([]), "Metadata is not empty")

    def test_create_metadata(self, namespace):
        # Given("Create namespace")
        # When ("Put metadata to namespace")
        key, value = 'key', 'value'
        put_metadata(namespace, key, value)
        # Then ("Check that metadata was added")
        meta_list = get_metadata_keys(namespace)
        assert_that(meta_list, has_length(1), "Metadata is not created")
        assert_that(meta_list, has_item(key), "Metadata is not created")

    def test_get_metadata_by_key(self, namespace, metadata):
        # Given ("Put metadata to namespace")
        meta_key, meta_value = metadata
        # When ("Get metadata by key")
        value = get_metadata_by_key(namespace, meta_key)
        # Then ("Check that metadata was added")
        assert_that(value, equal_to(meta_value), "Can't get meta value by key")

    def test_get_metadata_keys(self, namespace, metadata):
        # Given ("Put metadata to namespace")
        meta_key, meta_value = metadata
        # When ("Get list of metadata keys")
        meta_list = get_metadata_keys(namespace)
        # Then ("Check that metadata was added")
        assert_that(meta_list, has_item(meta_key), "Can't get list of meta keys")

    def test_update_metadata(self, namespace, metadata):
        # Given ("Put metadata to namespace")
        meta_key, meta_value = metadata
        # When ("Update metadata")
        new_meta_value = 'new_value'
        put_metadata(namespace, meta_key, new_meta_value)
        # Then ("Check that metadata was updated")
        updated_value = get_metadata_by_key(namespace, meta_key)
        assert_that(updated_value, equal_to(new_meta_value), "Can't update metadata")



from tests.helpers.log_helper import log_operation


def sql_query(namespace, query):
    db, namespace_name = namespace
    item_list = list(db.select(query))
    log_operation.info(f"Execute SQL query: {query}, got result: {item_list}")
    return item_list



from hamcrest import *

from tests.helpers.namespace import *


class TestCrudNamespace:
    def test_create_ns(self, database):
        # Given("Create namespace in empty database")
        namespace_name = 'test_ns'
        create_namespace(database, namespace_name)
        # When ("Get namespaces list in created database")
        namespace_list = get_namespace_list(database)
        # Then ("Check that database contains created namespace")
        assert_that(namespace_list, has_item(has_entry("name", equal_to(namespace_name))),
                    "Namespace wasn't created")
        drop_namespace(database, namespace_name)

    def test_delete_ns(self, database):
        # Given("Create namespace")
        namespace_name = 'test_ns'
        create_namespace(database, namespace_name)
        # When ("Delete namespace")
        drop_namespace(database, namespace_name)
        # Then ("Check that namespace was deleted")
        namespace_list = get_namespace_list(database)
        assert_that(namespace_list, not_(has_item(has_entry("name", equal_to(namespace_name)))),
                    "Namespace wasn't deleted")

    def test_cannot_delete_ns_not_created(self, database):
        # Given ("Create empty database")
        # Then ("Check that we cannot delete namespace that does not exist")
        namespace_name = 'test_ns'
        assert_that(calling(drop_namespace).with_args(database, namespace_name), raises(Exception, matching=has_string(
            f"Namespace '{namespace_name}' does not exist")))





from hamcrest import *

from tests.helpers.sql import sql_query


class TestSqlQueries:
    def test_sql_select(self, namespace, index, item):
        # Given("Create namespace with item")
        db, namespace_name = namespace
        item_definition = item
        # When ("Execute SQL query SELECT")
        query = f'SELECT * FROM {namespace_name}'
        item_list = sql_query(namespace, query)
        # Then ("Check that selected item is in result")
        assert_that(item_list, has_item(equal_to(item_definition)), "Can't SQL select data")

    def test_sql_select_with_join(self, namespace, second_namespace_for_join, index, items):
        # Given("Create two namespaces")
        db, namespace_name = namespace
        second_namespace_name, second_ns_item_definition_join = second_namespace_for_join
        # When ("Execute SQL query SELECT with JOIN")
        query = f'SELECT id FROM {namespace_name} INNER JOIN {second_namespace_name} ON {namespace_name}.id = {second_namespace_name}.id'
        item_list = sql_query(namespace, query)
        # Then ("Check that selected item is in result")
        assert_that(item_list,
                    has_item(equal_to({'id': 1, f'joined_{second_namespace_name}': [second_ns_item_definition_join]})),
                    "Can't SQL select data with JOIN")

    def test_sql_select_with_condition(self, namespace, index, items):
        # Given("Create namespace with item")
        db, namespace_name = namespace
        # When ("Execute SQL query SELECT")
        query = f'SELECT * FROM {namespace_name} WHERE id=3'
        item_list = sql_query(namespace, query)
        # Then ("Check that selected item is in result")
        assert_that(item_list, has_item(equal_to({'id': 3, 'val': 'testval3'})), "Can't SQL select data with condition")

    def test_sql_update(self, namespace, index, item):
        # Given("Create namespace with item")
        db, namespace_name = namespace
        # When ("Execute SQL query UPDATE")
        query = f"UPDATE {namespace_name} SET \"val\" = 'new_val' WHERE id = 100"
        item_list = sql_query(namespace, query)
        # Then ("Check that item is updated")
        assert_that(item_list, has_item(equal_to({'id': 100, 'val': 'new_val'})), "Can't SQL update data")

    def test_sql_delete(self, namespace, index, item):
        # Given("Create namespace with item")
        db, namespace_name = namespace
        # When ("Execute SQL query DELETE")
        query_delete = f"DELETE FROM {namespace_name} WHERE id = 100"
        sql_query(namespace, query_delete)
        # Then ("Check that item is deleted")
        query_select = f"SELECT * FROM {namespace_name}"
        item_list = sql_query(namespace, query_select)
        assert_that(item_list, equal_to([]), "Can't SQL delete data")

    def test_sql_select_with_syntax_error(self, namespace, index, item):
        # Given("Create namespace with item")
        # When ("Execute SQL query SELECT with incorrect syntax")
        query = f'SELECT *'
        # Then ("Check that selected item is in result")
        assert_that(calling(sql_query).with_args(namespace, query),
                    raises(Exception, matching=has_string(string_contains_in_order(
                        "Expected", "but found"))), "Error wasn't raised when syntax was incorrect")



from hamcrest import *

from tests.helpers.index import *
from tests.helpers.namespace import get_ns_description
from tests.test_data.constants import index_definition, updated_index_definition


class TestCrudIndexes:
    def test_initial_namespace_has_no_indexes(self, database, namespace):
        # Given("Create namespace")
        # When ("Get namespace information")
        ns_entry = get_ns_description(database, namespace)
        # Then ("Check that list of indexes in namespace is empty")
        assert_that(ns_entry, has_item(has_entry("indexes", equal_to([]))),
                    "Index is not empty")

    def test_create_index(self, database, namespace):
        # Given("Create namespace")
        # When ("Add index")
        create_index(namespace, index_definition)
        # Then ("Check that index is added")
        ns_entry = get_ns_description(database, namespace)
        assert_that(ns_entry, has_item(has_entry("indexes", has_item(equal_to(index_definition)))),
                    "Index wasn't created")
        drop_index(namespace, 'id')

    def test_update_index(self, database, namespace, index):
        # Given("Create namespace with index")
        # When ("Update index")
        update_index(namespace, updated_index_definition)
        # Then ("Check that index is updated")
        ns_entry = get_ns_description(database, namespace)
        assert_that(ns_entry, has_item(has_entry("indexes", has_item(equal_to(updated_index_definition)))),
                    "Index wasn't updated")

    def test_delete_index(self, database, namespace):
        # Given("Create namespace with index")
        create_index(namespace, index_definition)
        # When ("Delete index")
        drop_index(namespace, 'id')
        # Then ("Check that index is deleted")
        ns_entry = get_ns_description(database, namespace)
        assert_that(ns_entry, has_item(has_entry("indexes", equal_to([]))),
                    "Index wasn't deleted")

    def test_cannot_add_index_with_same_name(self, database, namespace):
        # Given("Create namespace")
        # When ("Add index")
        create_index(namespace, index_definition)
        # Then ("Check that we can't add index with the same name")
        assert_that(calling(create_index).with_args(namespace, updated_index_definition),
                    raises(Exception, matching=has_string(string_contains_in_order(
                        "Index", "already exists with different settings"))),
                    "Index with existing name was created")

    def test_cannot_update_not_existing_index_in_namespace(self, database, namespace):
        # Given ("Create namespace")
        db, namespace_name = namespace
        # When ("Update index")
        # Then ("Check that we can't update index that was not created")
        assert_that(calling(update_index).with_args(namespace, index_definition),
                    raises(Exception, matching=has_string(f"Index 'id' not found in '{namespace_name}'")),
                    "Not existing index was updated")

    def test_cannot_delete_not_existing_index_in_namespace(self, database, namespace):
        # Given ("Create namespace")
        # When ("Delete index")
        # Then ("Check that we can't delete index that was not created")
        index_name = 'id'
        assert_that(calling(drop_index).with_args(namespace, index_name),
                    raises(Exception, matching=has_string(f"Cannot remove index {index_name}: doesn't exist")),
                    "Not existing index was deleted")



from hamcrest import *

from tests.helpers.index import create_index
from tests.helpers.items import *
from tests.test_data.constants import item_definition


class TestCrudItems:
    def test_initial_namespace_has_no_items(self, namespace, index):
        # Given("Create namespace with index")
        db, namespace_name = namespace
        # When ("Get namespace information")
        select_result = list(db.select(f'SELECT * FROM {namespace_name}'))
        # Then ("Check that list of items in namespace is empty")
        assert_that(select_result, has_length(0), "Item list is not empty")
        assert_that(select_result, equal_to([]), "Item list is not empty")

    def test_create_item_insert(self, namespace, index):
        # Given("Create namespace with index")
        db, namespace_name = namespace
        # When ("Insert item into namespace")
        insert_item(namespace, item_definition)
        # Then ("Check that item is added")
        select_result = list(db.select(f'SELECT * FROM {namespace_name}'))
        assert_that(select_result, has_length(1),
                    "Item wasn't created")
        assert_that(select_result, has_item(item_definition),
                    "Item wasn't created"
                    )
        delete_item(namespace, item_definition)

    def test_create_item_insert_with_precepts(self, namespace, index):
        # Given("Create namespace with index")
        db, namespace_name = namespace
        # When ("Insert items into namespace")
        number_items = 5
        for _ in range(number_items):
            db.item_insert(namespace_name, {"id": 100, "field": "value"}, ["id=serial()"])
        # Then ("Check that item is added")
        select_result = list(db.select(f'SELECT * FROM {namespace_name}'))
        assert_that(select_result, has_length(number_items),
                    "Items wasn't created")
        for i in range(number_items):
            assert_that(select_result[i], equal_to({'id': i+1, "field": "value"}),
                    "Items wasn't created")
        for i in range(number_items):
            db.item_delete(namespace_name, {'id': i})

    def test_create_item_upsert(self, namespace, index):
        # Given("Create namespace with index")
        db, namespace_name = namespace
        # When ("Upsert item into namespace")
        upsert_item(namespace, item_definition)
        # Then ("Check that item is added")
        select_result = list(db.select(f'SELECT * FROM {namespace_name}'))
        assert_that(select_result, has_length(1),
                    "Item wasn't created")
        assert_that(select_result, has_item(item_definition),
                    "Item wasn't created")
        delete_item(namespace, item_definition)

    def test_update_item_upsert(self, namespace, index, item):
        # Given("Create namespace with item")
        db, namespace_name = namespace
        # When ("Upsert item")
        item_definition_updated = {'id': 100, 'val': "new_value"}
        upsert_item(namespace, item_definition_updated)
        # Then ("Check that item is updated")
        select_result = list(db.select(f'SELECT * FROM {namespace_name}'))
        assert_that(select_result, has_length(1),
                    "Item wasn't updated")
        assert_that(select_result, has_item(item_definition_updated),
                    "Item wasn't updated")

    def test_update_item_update(self, namespace, index, item):
        # Given("Create namespace with item")
        db, namespace_name = namespace
        # When ("Update item")
        item_definition_updated = {'id': 100, 'val': "new_value"}
        update_item(namespace, item_definition_updated)
        # Then ("Check that item is updated")
        select_result = list(db.select(f'SELECT * FROM {namespace_name}'))
        assert_that(select_result, has_length(1), "Item wasn't updated")
        assert_that(select_result, has_item(item_definition_updated), "Item wasn't updated")

    def test_delete_item(self, namespace, index):
        # Given("Create namespace with item")
        db, namespace_name = namespace
        insert_item(namespace, item_definition)
        # When ("Delete item")
        delete_item(namespace, item_definition)
        # Then ("Check that item is deleted")
        select_result = list(db.select(f'SELECT * FROM {namespace_name}'))
        assert_that(select_result, has_length(0), "Item wasn't deleted")
        assert_that(select_result, equal_to([]), "Item wasn't deleted")



import logging

from tests.helpers.log_helper import log_operation


def create_namespace(database, namespace_name):
    """
    Create a namespace
    """
    db, db_name = database
    log_operation.info(f"Create a namespace with name '{namespace_name}' on database '{db_name}'")
    try:
        db.namespace_open(namespace_name)
    except Exception as e:
        logging.error(e)


def drop_namespace(database, namespace_name):
    """
    Drop a namespace
    """
    db, db_name = database
    log_operation.info(f"Drop a namespace with name '{namespace_name}' on database '{db_name}'")
    db.namespace_drop(namespace_name)


def get_namespace_list(database):
    """
    Get list of namespaces in database
    """
    log_operation.info("Get list of namespaces in database")
    db, db_name = database
    namespace_list = db.namespaces_enum()
    return namespace_list


def get_ns_description(database, namespace):
    """
    Get information about namespace in database
    """
    db, namespace_name = namespace
    namespace_list = get_namespace_list(database)
    log_operation.info(f"Get information about namespace {namespace_name} in database")
    ns_entry = list(filter(lambda ns: ns['name'] == namespace_name, namespace_list))
    return ns_entry



from tests.helpers.log_helper import log_operation


def insert_item(namespace, item_def):
    """
    Insert item to namespace
    """
    db, namespace_name = namespace
    log_operation.info(f"Insert item: {item_def} to namespace {namespace_name}")
    db.item_insert(namespace_name, item_def)


def upsert_item(namespace, item_def):
    """
    Insert or update item to namespace
    """
    db, namespace_name = namespace
    log_operation.info(f"Upsert item: {item_def} to namespace {namespace_name}")
    db.item_upsert(namespace_name, item_def)


def update_item(namespace, item_def):
    """
    Update item to namespace
    """
    db, namespace_name = namespace
    log_operation.info(f"Update item: {item_def} to namespace {namespace_name}")
    db.item_upsert(namespace_name, item_def)


def delete_item(namespace, item_def):
    """
    Delete item from namespace
    """
    db, namespace_name = namespace
    log_operation.info(f"Delete item: {item_def} from namespace {namespace_name}")
    db.item_delete(namespace_name, item_def)



import pyreindexer
from hamcrest import *

from tests.test_data.constants import special_namespaces, special_namespaces_cluster


class TestCrudDb:
    def test_create_db(self):
        # Given ("Create empty database")
        db_path = '/tmp/test_db'
        db = pyreindexer.RxConnector('builtin://' + db_path)
        # When ("Get namespaces list in created database")
        namespace_list = db.namespaces_enum()
        # Then ("Check that database contains only special namespaces")
        expect_namespaces = special_namespaces
        if len(namespace_list) == len(special_namespaces_cluster):
            expect_namespaces = special_namespaces_cluster
        for namespace in expect_namespaces:
            assert_that(namespace_list, has_item(has_entries("name", equal_to(namespace["name"]))),
                        "Database doesn't contain special namespaces")
        assert_that(namespace_list, has_length(equal_to(len(expect_namespaces))),
                    "Database doesn't contain special namespaces")
